==> lincks-2.1.tar.gz: lincks-2.1/interface/xlincks/pen_notify.c <==
/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * MODULE NAME: 	pen_notify.c
 *
 * SCCSINFO:		@(#)pen_notify.c	1.3 12/15/93
 *
 * ORIGINAL AUTHOR(S):  Anders Broman,pumgruppen PEN 1992-05-28
 *
 * MODIFICATIONS:
 *	1993-08-12 Martin Sjlin - changed liblincks interface
 *              as we retrieve all possible pens at once ....
 *      1993-11-02 Martin Sjlin - made the setup in PEN_oob do a
 *              callback to user interface level. Removed reference
 *              to toplevel and all X stuff. Moved into xstuff.c
 *	<list mods here with name and date>
 *      
 * DESCRIPTION:
 *
 * This module creates and displays a warningsmessage when
 * parallel editing occurs.
 * It uses X window routines to build the message window.
 *
 *********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************
 */
int is_penned( /* int node */);
int clear_penned(/* int node */);
void highlight_items();
void PEN_oob( /* char oob */ );

/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include <unistd.h>
#include <sys/ioctl.h>
#include <stdio.h>
#if !defined(FIONREAD) && defined(SVR4)
#include <sys/filio.h>
#endif /* n FIONREAD && SVR4 */
#ifdef SCO
#include <sys/socket.h>
#endif
#ifdef NO_FIONREAD_PIPE
#include <errno.h>
#include <fcntl.h>
#endif /* NO_FIONREAD_PIPE */

#ifndef sequent
#include <malloc.h>
#endif	/* sequent */
#include "aimtypes.h"
#include "systypes.h"

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
/* xstuff.c */
extern void popup_message( /* varargs */);
extern void short_error(/* varargs */);
extern void uim_highlight( /* infonode *inp,int how */ );
extern int  pen_select_func( /* int fd, void (* pen_call_back)() */);

/* usfuns.c */
extern errcode PEN_RETRIEVE_MESSAGE( /* int (*mapfm)(), char *extra */ );

/* libshared */
extern char *strdup( /* char *incoming */ );

#ifdef sequent
extern char *malloc();
extern void free();
#endif  /* sequent */

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern reference_structure *refstructs; 	/* aimstart.c */
extern int suppress;				/* aimstart.c */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
typedef struct _penreceived { /* a pen message that's been received */
  int type;
  int index;
  char *message;
  struct _penreceived *next;
} penreceived;

#define MSGSIZE 1024

/* Copied from liblincks/dbcodes.h */
#define OOB_MESSAGE_CODE 0x1  /* pen message */
#define PEN_RESOLVE     (0) /* Resolve version node async. */
#define PEN_CHECK       (1) /* Node being edited by other user */
#define PEN_POLL        (2) /* User began editing in parallel */


/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static void pen_notify();
static int is_bt( /* int object */ );
static int is_new_pen( /* int index, int type, char *msg */);
static int add_new_pen(/* int index, int type, char *msg */);
static void highlight_identry( /* identry *idp */ );
static void highlight_infonode( /* infonode *inp*/ );
static int  PEN_mapfn( /* char *extra, int index, int type, 
			      char *msg, int count, int len */ );

/*********************************************************************
 * INTERNAL (STATIC) DATA: 
 *********************************************************************/
static int PEN_pipe[2];
static int PEN_inited = 0;
static penreceived *penlist = NULL;
static char message[MSGSIZE]; 

/*  */
/**********************************************************************
 * Function: static void pen_notify()
 * 
 * present the message on the screen
 *
 * Modifications:
 *      1993-08-12 Martin Sjlin. Now conforms to new lowlevel
 *                 liblincks interface for OOB/PENS.
 */
static void pen_notify()
{
  long pens,size = 0;
  int  count;
  int  st;
#ifdef NO_FIONREAD_PIPE
  int  flags;
#endif /* NO_FIONREAD_PIPE */

#ifdef NO_FIONREAD_PIPE
  /* It seems like you cannot do a FIONREAD on a pipe under SCO! */ 
  if ((flags = fcntl(PEN_pipe[0], F_GETFL, 0)) < 0) {
    perror("pen_notify: fcntl(1) failed to get on pipe");
    return;
  }
# ifdef O_NONBLOCK
  if (fcntl(PEN_pipe[0], F_SETFL, flags | O_NONBLOCK) < 0) {
# else
  if (fcntl(PEN_pipe[0], F_SETFL, flags | O_NDELAY) < 0) {
# endif /* n O_NONBLOCK */
    perror("pen_notify: fcntl(2) failed to set on pipe");
    return;
  }
  size = 64;				/* try some large number */
#else
  /* we get number character available in pipe and read/skip them */
  if ( ioctl(PEN_pipe[0],FIONREAD,&size) < 0) 
    perror("pen_notify: ioctl failed");
#ifdef PENDEBUG
  fprintf( stderr, "pen_notify: %d characters in pipe\n", size);
#endif
  if (size <= 0)
    return;
#endif /* n NO_FIONREAD_PIPE */

  /* OK, skip number of character in pipe */
  for(pens = 0; pens < size; pens += count) {
    count = size < MSGSIZE ? size : MSGSIZE;
#ifdef NO_FIONREAD_PIPE
    if ((count = read( PEN_pipe[0], message, count)) > 0)
      continue;
    else
# ifdef O_NONBLOCK
    if (count == -1 && errno == EAGAIN) break;
# else /* to handle SysV(O_NDELAY) (0) and 4.2BSD(O_NDELAY) (-1) */
    if (count <=  0 && errno == EWOULDBLOCK) break;
# endif /* n O_NONBLOCK */
    else {
      perror("pen_notify: non-blocking read on pipe failed");
      goto exit;
    } 
#else /* NO_FIONREAD_PIPE */
    if ( read( PEN_pipe[0], message, count) != count ) {
      perror("pen_notify: read failed");
      return;
    }
#endif /* n NO_FIONREAD_PIPE */
  } /* for */

#if defined(PENDEBUG) && defined(NO_FIONREAD_PIPE)
  fprintf( stderr, "pen_notify: %d characters in pipe\n", pens);
#endif

  if ( (st = PEN_RETRIEVE_MESSAGE(PEN_mapfn, (char *)&pens)) != SUCCESS )
    fprintf(stderr, "pen_notify: Can't retrieve/treat messages (%d)\n", st);
#ifdef NO_FIONREAD_PIPE
 exit:
  if (fcntl(PEN_pipe[0], F_SETFL, flags) < 0) 
    perror("pen_notify: fcntl(3) failed to set on pipe");
#endif /* NO_FIONREAD_PIPE */

}

/*  */
/**********************************************************************
 * Function: static int PEN_mapfn(char *extra, int index, int type, char *msg, int count, int len)
 *
 * Is called once for every pen message - for parameter description
 * see usfuns.c. Is always called at least once, for setup. Returns
 * nonzero value and it will no longer be call during processing 
 *
 */
static int PEN_mapfn( extra, index, type, msg, count, len)
 char *extra;
 int   index;
 int   type;
 char *msg; 
 int   count; 
 int   len;
{
  static char *object_kind[] = { "object",
				 "binding table" };

#ifdef PENDEBUG 
  fprintf( stderr,"PEN_mapfn: index %d type %d count %d len %d msg %s\n",
	   index, type, count, len, msg != NULL ? msg : "");
#endif
  /* no init is needed here, we skip the first time */
  if ( count == 0 ) 
    ;
  /* The history node msg.index has been merged from a parallel update.  */
  else if (type == PEN_RESOLVE) 
    (void)printf("The history node %d was merged from a parallel update.\n",
		 index);
  else if (!msg ) 
    short_error("message NULL in pen_notify.\n");
  else if (is_new_pen( index, type, msg)) {
    int isbt = is_bt(index);

    /* if we fail to add this message, n.c. - we only see it once more */
    (void) add_new_pen( index, type, msg);

    bzero(message, sizeof(message));

    switch (type) {
    case PEN_CHECK:
      (void)sprintf(message, 
		    "The %s <%d:-1> is\nbeing edited by the user(s):\n %s",
		    object_kind[isbt], index, msg);
      break;

    case PEN_POLL:
      (void)sprintf(message, 
		    "User %s just began editing %s <%d:-1>.", 
		    msg, object_kind[isbt], index);
      break;

    default:
      /* error_message */
      perror("pen_notify: Unknown pen message type\n");
      break;

    } /* switch */

    if (!suppress || isbt)
      popup_message(POPUP_MESS_PEN, message);
    if (!isbt)
      highlight_items();
  } /* if is new pen */
    
  /* allocated in rdwrmol.c in ReceiveImage */
  if ( msg )
    (void) free(msg); 

  /* we want to see them all ... */
  return(0);
}

 
/*  */
/**********************************************************************
 * Function: void PEN_oob( char oob )
 * 
 * put an event in the event_queue
 *
 * Modifications:
 *      1993-11-02 Martin Sjlin. Make a 'callback' to an X function
 *                 or whatever we have as an userinterface. To isolate
 *                 the aim layer further from user interface layers.
 *      <list mods with name and date>
 */
void PEN_oob(oob)
char oob;
{
  switch( PEN_inited ) {
  case 0:
    if (pipe(PEN_pipe) != 0) {
      perror("PEN_oob: Can't open pipe\n");
      break;
    }
    PEN_inited = 1;

  case 1:
    /* 
     * the call back is called with the file descriptor, and 
     * the function to be called when we get something
     * on the file descriptor (e.g in X do a XtAppAddInput ... )
     */
    if (pen_select_func(PEN_pipe[0],pen_notify)) 
      PEN_inited = 2;

  case 2:
    if( write(PEN_pipe[1], &oob, 1) != 1) 
      perror("PEN_oob: Write failed");
#ifdef PENDEBUG
    fprintf( stderr, "PEN_oob: Wrote pen %d on pipe\n", (int) oob);
#endif
    break;
  }
}

/*  */
/**********************************************************************
 * Function: static int is_new_pen(int index, int type, char *msg)
 * 
 * Return 0 if the given node is in the penlist, otherwise 1.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int is_new_pen( index, type, msg )
 int index;
 int type;
 char *msg;
{
  penreceived *p = penlist;

  for (p = penlist; p != NULL; p = p->next) 
    if ( p->index == index &&
	 p->type == type    &&
         strcmp(p->message,msg) == 0) 
      return(0);

  return(1);
}

/*  */
/**********************************************************************
 * Function: static int add_new_pen(int index, int type, char *msg)
 * 
 * Add the new pen to the front of the list ... return 1 if no error
 * else 0
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int add_new_pen( index, type, msg)
 int index;
 int type;
 char *msg;
{
  penreceived *p = penlist;

  p = (penreceived *)malloc(sizeof(penreceived));
  if ( !p )
    return(0);
  
  p->index   = index;
  p->type    = type;
  p->message = strdup(msg);
  p->next    = penlist;
  penlist    = p;
  return(1);
}

/*  */
/**********************************************************************
 * Function: int is_penned(int node)
 * 
 * Return 1 if the given node is in the penlist, otherwise 0.
 *
 * Modifications:
 *      <list mods with name and date>
 */
int is_penned(node)
  int node;
{
  penreceived *tmp;

  for (tmp = penlist; tmp != NULL; tmp = tmp->next) {
    if (tmp->index == node)
      return 1;
  }
  return 0;
}

/*  */
/**********************************************************************
 * Function: int clear_penned(int node)
 * 
 * Removes node from the penlist. Return 1 if it was in the list,
 * otherwise 0.
 *
 * Modifications:
 *      <list mods with name and date>
 */
int clear_penned(node)
  int node;
{
  penreceived *tmp, *p = NULL;

  for (tmp = penlist; tmp != NULL; tmp = tmp->next) {
    if (tmp->index == node) {
      if (p) {
	p->next = tmp->next;
      } else {
	penlist = tmp->next;
      }
      if (tmp->message)
	free(tmp->message);
      free((char *)tmp);
      return 1;
    }
  }
  return 0;
}


/*  */
/**********************************************************************
 * Function: static int is_bt(int object)
 * 
 * return whether an object is used as a binding table or not.  note 
 * that this function isn't close to infallible.  it will not find a
 * binding table on a closed but unsaved reference structure.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int is_bt(object)
  int object;
{
  reference_structure *ref;

  for (ref = refstructs; ref != NULL; ref = ref -> next) {
    if (ref->bt.vs == object)
      return 1;
  }
  return 0;
}

/*  */
/**********************************************************************
 * Function: void highlight_items()
 * 
 * Modifications:
 *      <list mods with name and date>
 */
void highlight_items()
{
  reference_structure *ref;

  for (ref = refstructs; ref != NULL; ref = ref -> next) {
    highlight_identry(ref->target);
  }
}

/*  */
/**********************************************************************
 * Function: static void highlight_identry(identry *idp)
 * 
 * Modifications:
 *      <list mods with name and date>
 */
static void highlight_identry(idp)
  identry *idp;
{
  while (idp != NULL) {
    highlight_infonode(idp->value);
    idp = idp -> next; 
  }
}

/*  */
/**********************************************************************
 * Function: static void highlight_infonode(infonode *inp)
 * 
 * Modifications:
 *      <list mods with name and date>
 */
static void highlight_infonode(inp)
  infonode *inp;
{
  while (inp != NULL) {
    if (inp -> subnodes)
      highlight_identry(inp->subnodes);
    else
      uim_highlight(inp);
    inp = inp -> next; 
  }
}

==> lincks-2.2.1.tar.gz: lincks-2.2.1/interface/xlincks/pen_notify.c <==
/* 
 * Linkoping Intelligent Communication of Knowledge System (LINCKS)
 *      Copyright (C) 1993, 1994 Lin Padgham, Ralph Rnnquist
 *       Department of Computer and Information Sciences
 *		University of Linkoping, Sweden
 *		    581 83 Linkoping, Sweden
 *		       lincks@ida.liu.se
 *
 * These collective LINCKS programs are free software; you can 
 * redistribute them and/or modify them under the terms of the GNU
 * General Public License as published by the Free Software Foundation,
 * version 2 of the License.
 *
 * These programs are distributed in the hope that they will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the programs; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* 
 * MODULE NAME: 	pen_notify.c
 *
 * SCCSINFO:		@(#)pen_notify.c	1.10 5/20/94
 *
 * ORIGINAL AUTHOR(S):  Anders Broman,pumgruppen PEN 1992-05-28
 *
 * MODIFICATIONS:
 *	1993-08-12 Martin Sjlin - changed liblincks interface
 *              as we retrieve all possible pens at once ....
 *      1993-11-02 Martin Sjlin - made the setup in PEN_oob do a
 *              callback to user interface level. Removed reference
 *              to toplevel and all X stuff. Moved into xstuff.c
 *      1994-01-07 Martin Sjlin - followup the changes to the call
 *              interface (unified with RCB).
 *      1994-01-13 Martin Sjlin - rewrote callback for PEN and thus
 *              removed all lowlevel things from this module.
 *	<list mods here with name and date>
 *      
 * DESCRIPTION:
 *
 * This module creates and displays a warningsmessage when
 * parallel editing occurs.
 * It uses X window routines to build the message window.
 */
/*********************************************************************
 * INCLUDES:
 *********************************************************************/
#include "config.h"	/* includes system dependent includes */
#include "aimtypes.h"

/*********************************************************************
 * EXTERNALLY-AVAILABLE	DATA FOUND IN THIS MODULE:
 *********************************************************************/
/* none */

/*********************************************************************
 * EXTERNALLY-CALLABLE ROUTINES FOUND IN THIS MODULE:
 *********************************************************************/
#include "f_pen_notify.h"

/*********************************************************************
 * EXTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
#include "f_xstuff.h"
#include "f_chng_prop.h"		/* mark_reference */

/* libshared */
extern char *strdup( /* char *incoming */ );

/*********************************************************************
 * EXTERNAL DATA STRUCTURES USED BY THIS MODULE:
 *********************************************************************/
extern reference_structure *refstructs; 	/* aimstart.c */
extern int suppress;				/* aimstart.c */

/*********************************************************************
 * LOCAL DEFINES, STRUCTS, TYPEDEFS, ETC.:
 *********************************************************************/
typedef struct _penreceived { /* a pen message that's been received */
#define POLL  1
#define CHECK 2
  int type;
  int index;
  char *message;
  struct _penreceived *next;
} penreceived;

/*********************************************************************
 * INTERNAL FUNCTIONS USED BY THIS MODULE:
 *********************************************************************/
static int add_new_pen P_(( int index, int type, char *msg ));
static int is_bt P_(( int object ));
static int is_new_pen P_(( int index, int type, char *msg ));

/*********************************************************************
 * INTERNAL (STATIC) DATA: 
 *********************************************************************/
static char *object_kind[] = { "object", "binding table" };
static penreceived *penlist = NULL;

/*  */
/**********************************************************************
 * int PEN_resolve(void *extra, int index, char *msg);
 *
 * PEN handler for RESOLVE class messages.
 *
 * Modifications:
 *      <list mods with name and date>
 */
int PEN_resolve(extra, index,  msg)
  void *extra;
  int   index;
  char *msg;
{
  (void)printf("The history node %d was merged from a parallel update.\n",
		 index);
  return(1);
}

/*  */
/**********************************************************************
 * int PEN_check(void *extra, int index, char *msg);
 *
 * PEN handler for CHECK class messages.
 *
 * Modifications:
 *      <list mods with name and date>
 */
int PEN_check(extra, index, msg)
  void *extra;
  int   index;
  char *msg;
{
  if (!msg ) 
    short_error("message NULL in PEN_check.\n");
  else if (is_new_pen( index, CHECK, msg)) {
    int isbt = is_bt(index);

    /* if we fail to add this message, n.c. - we only see it once more */
    (void) add_new_pen( index, CHECK, msg);
    if (!suppress || isbt)
      popup_message(POPUP_MESS_PEN,
		    "The %s <%d:-1> is\nbeing edited by the user(s):\n %s",
		    object_kind[isbt], index, msg);
    if (!isbt)
      mark_reference(refstructs, uim_highlight);

  } /* is_new_pen */

  return(1);
}

/*  */
/**********************************************************************
 * int PEN_poll(void *extra, int index, char *msg);
 *
 * PEN handler for POLL class messages.
 *
 * Modifications:
 *      <list mods with name and date>
 */
int PEN_poll(extra, index, msg)
  void *extra;
  int   index;
  char *msg;
{
  if (!msg ) 
    short_error("message NULL in PEN_poll.\n");
  else if (is_new_pen( index, POLL, msg)) {
    int isbt = is_bt(index);

    /* if we fail to add this message, n.c. - we only see it once more */
    (void) add_new_pen( index, POLL, msg);
    if (!suppress || isbt)
      popup_message(POPUP_MESS_PEN, 
		   "User %s just began editing %s <%d:-1>.", 
		   msg, object_kind[isbt], index);
    if (!isbt)
      mark_reference(refstructs, uim_highlight);

  } /* is_new_pen */

  return(1);
}
 
/*  */
/**********************************************************************
 * Function: static int is_new_pen(int index, int type, char *msg)
 * 
 * Return 0 if the given node is in the penlist, otherwise 1.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int is_new_pen( index, type, msg )
 int index;
 int type;
 char *msg;
{
  penreceived *p = penlist;

  for (p = penlist; p != NULL; p = p->next) 
    if ( p->index == index &&
	 p->type == type    &&
         strcmp(p->message,msg) == 0) 
      return(0);

  return(1);
}

/*  */
/**********************************************************************
 * Function: static int add_new_pen(int index, int type, char *msg)
 * 
 * Add the new pen to the front of the list ... return 1 if no error
 * else 0
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int add_new_pen( index, type, msg)
 int index;
 int type;
 char *msg;
{
  penreceived *p = penlist;

  p = (penreceived *)malloc(sizeof(penreceived));
  if ( !p )
    return(0);
  
  p->index   = index;
  p->type    = type;
  p->message = strdup(msg);
  p->next    = penlist;
  penlist    = p;
  return(1);
}

/*  */
/**********************************************************************
 * Function: int is_penned(int node)
 * 
 * Return 1 if the given node is in the penlist, otherwise 0.
 *
 * Modifications:
 *      <list mods with name and date>
 */
int is_penned(node)
  int node;
{
  penreceived *tmp;

  for (tmp = penlist; tmp != NULL; tmp = tmp->next) {
    if (tmp->index == node)
      return 1;
  }
  return 0;
}

/*  */
/**********************************************************************
 * Function: int clear_penned(int node)
 * 
 * Removes node from the penlist. Return 1 if it was in the list,
 * otherwise 0.
 *
 * Modifications:
 *      <list mods with name and date>
 */
int clear_penned(node)
  int node;
{
  penreceived *tmp, *p = NULL;

  for (tmp = penlist; tmp != NULL; tmp = tmp->next) {
    if (tmp->index == node) {
      if (p) {
	p->next = tmp->next;
      } else {
	penlist = tmp->next;
      }
      if (tmp->message)
	free(tmp->message);
      free((FREEPTR *)tmp);
      return 1;
    }
  }
  return 0;
}

/*  */
/**********************************************************************
 * Function: static int is_bt(int object)
 * 
 * return whether an object is used as a binding table or not.  note 
 * that this function isn't close to infallible.  it will not find a
 * binding table on a closed but unsaved reference structure.
 *
 * Modifications:
 *      <list mods with name and date>
 */
static int is_bt(object)
  int object;
{
  reference_structure *ref;

  for (ref = refstructs; ref != NULL; ref = ref -> next) {
    if (ref->bt.vs == object)
      return 1;
  }
  return 0;
}
